==> scaling_vin_pytorch-0.0.5-py3-none-any.whl: scaling_vin_pytorch/vin.py <==
from functools import partial, wraps

import torch
import torch.nn.functional as F
from torch.nn import Module, ModuleList
from torch import nn, tensor, Tensor, is_tensor
from torch.utils.checkpoint import checkpoint_sequential

from x_transformers import Attention
from x_transformers.x_transformers import RotaryEmbedding

import einx
from einops import rearrange, repeat, reduce, einsum, pack, unpack

# ein notation
# b - batch
# c - channels
# s - state dimension
# v - values
# a - actions (channels)
# o - output (channels)
# i - input (channels)
# h - height
# w - width
# d - depth of value iteration network
# p - number of plans (different value maps transitioned on)

# helpers

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

def divisible_by(num, den):
    return (num % den) == 0

def is_odd(n):
    return not divisible_by(n, 2)

def pack_one(t, pattern):
    output, packed_shape = pack([t], pattern)

    def inverse_fn(packed, override_pattern = None):
        override_pattern = default(override_pattern, pattern)

        return unpack(packed, packed_shape, override_pattern)[0]

    return output, inverse_fn

def should_checkpoint(
    self,
    inputs: Tensor | tuple[Tensor, ...],
    check_instance_variable = 'checkpoint'
) -> bool:
    if is_tensor(inputs):
        inputs = (inputs,)

    return (
        self.training and
        any([i.requires_grad for i in inputs]) and
        (not exists(check_instance_variable) or getattr(self, check_instance_variable, False))
    )

# tensor helpers

def soft_maxpool(
    t,
    temperature = 1.,
    dim = 1,
):
    t = t / temperature
    out = t.softmax(dim = dim) * t
    out = out.sum(dim = dim)
    return out * temperature

# modules and classes

class ActionSelector(Module):
    def __init__(
        self,
        num_plans = 1,
        soft_maxpool = False,
        soft_maxpool_temperature = 1,
    ):
        super().__init__()
        self.num_plans = num_plans

        # allow for softmax(x) * x pooling
        # https://mpflueger.github.io/assets/pdf/svin_iclr2018_v2.pdf
        self.soft_maxpool = soft_maxpool
        self.soft_maxpool_temperature = soft_maxpool_temperature

    def forward(self, q_values):

        q_values = rearrange(q_values, 'b (p a) ... -> b p a ...', p = self.num_plans)

        if not self.soft_maxpool:
            next_values = q_values.amax(dim = 2)
        else:
            next_values = soft_maxpool(q_values, temperature = self.soft_maxpool_temperature, dim = 2)

        return next_values

class ValueIteration(Module):
    def __init__(
        self,
        action_channels,
        *,
        num_plans = 1,
        receptive_field = 3,
        pad_value = 0.,
        soft_maxpool = False,
        soft_maxpool_temperature = 1.,
        dynamic_transition_kernel = True,
    ):
        super().__init__()
        assert is_odd(receptive_field)
        padding = receptive_field // 2

        self.action_channels = action_channels
        plan_actions = num_plans * action_channels

        conv_out = plan_actions * (1 if not dynamic_transition_kernel else receptive_field ** 2)
        self.transition = nn.Conv2d(num_plans, conv_out, receptive_field, padding = padding, bias = False)

        self.kernel_size = receptive_field

        self.pad_value = pad_value
        self.pad = partial(F.pad, pad = (padding,) * 4, value = pad_value)

        self.padding = padding

        self.action_selector = ActionSelector(
            num_plans = num_plans,
            soft_maxpool = soft_maxpool,
            soft_maxpool_temperature = soft_maxpool_temperature
        )

        self.dynamic_transition_kernel = dynamic_transition_kernel

    def forward(
        self,
        values,
        rewards
    ):
        pad = self.pad

        rewards_and_values = values + rewards

        # prepare for transition

        transition_weight = self.transition.weight

        # dynamic transition kernel - in other words, first convolution outputs the transition kernel

        if self.dynamic_transition_kernel:

            dynamic_transitions = F.conv2d(pad(rewards_and_values), transition_weight)

            # reshape the output into the next transition weight kernel

            dynamic_transitions = rearrange(dynamic_transitions, 'b (o i k1 k2) h w -> b o h w (i k1 k2)', k1 = self.kernel_size, k2 = self.kernel_size, o = self.action_channels)

            # the author then uses a softmax on the dynamic transition kernel
            # this actually becomes form of attention pooling seen in other papers

            dynamic_transitions = F.softmax(dynamic_transitions, dim = -1)

            # unfold the reward and values to manually do "conv" with data dependent kernel

            width = rewards_and_values.shape[-1] # for rearranging back after unfold

            unfolded_values = F.unfold(pad(rewards_and_values), self.kernel_size)
            unfolded_values = rearrange(unfolded_values, 'b i (h w) -> b i h w', w = width)

            # dynamic kernel

            q_values = einsum(unfolded_values, dynamic_transitions, 'b i h w, b o h w i -> b o h w')

        else:
            # transition

            q_values = F.conv2d(pad(rewards_and_values), transition_weight)

        # selecting the next action

        next_values = self.action_selector(q_values)

        return next_values

class ValueIterationNetwork(Module):
    def __init__(
        self,
        vi_module: ValueIteration,
        depth,
        checkpoint = True,
        checkpoint_segments = None
    ):
        super().__init__()
        self.depth = depth
        self.vi_module = vi_module

        self.checkpoint = checkpoint
        self.checkpoint_segments = checkpoint_segments

    def forward(
        self,
        values,
        rewards,
        depth = None
    ):
        depth = default(depth, self.depth)

        values, _ = pack_one(values, 'b * h w')
        rewards, _ = pack_one(rewards, 'b * h w')

        # checkpointable or not

        if depth > 1 and should_checkpoint(self, (values, rewards)):

            layer_values = [None] * depth

            segments = default(self.checkpoint_segments, depth)
            checkpoint_fn = partial(checkpoint_sequential, segments = segments, use_reentrant = False)

            def recurrent_layer(inputs):
                layer_ind, values, rewards, *layer_values = inputs

                next_values = self.vi_module(values, rewards)

                layer_values[layer_ind] = next_values

                return layer_ind + 1, next_values, rewards, *layer_values

            all_recurrent_layers = (recurrent_layer,) * depth

            _, _, _, *layer_values = checkpoint_fn(all_recurrent_layers, input = (0, values, rewards, *layer_values))

        else:

            layer_values = []

            for _ in range(depth):
                values = self.vi_module(values, rewards)
                layer_values.append(values)

        # return all values

        assert len(layer_values) == depth

        return layer_values

# main class

class ScalableVIN(Module):
    def __init__(
        self,
        state_dim,
        reward_dim,
        num_actions,
        num_plans = 1,
        dim_hidden = 150,
        init_kernel_size = 7,
        depth = 100,                    # they scaled this to 5000 in the paper to solve 100x100 maze
        loss_every_num_layers = 4,      # calculating loss every so many layers in the value iteration network
        final_cropout_kernel_size = 3,
        soft_maxpool = False,
        soft_maxpool_temperature = 1.,
        vi_module_kwargs: dict = dict(),
        vin_kwargs: dict = dict(),
        attn_dim_head = 64,
        attn_kwargs: dict = dict()
    ):
        super().__init__()
        self.depth = depth
        self.num_actions = num_actions
        self.num_plans = num_plans

        plan_actions = num_actions * num_plans

        self.reward_mapper = nn.Sequential(
            nn.Conv2d(state_dim + reward_dim, dim_hidden, 3, padding = 1),
            nn.Conv2d(dim_hidden, 1, 1, bias = False)
        )

        self.to_init_value = nn.Conv2d(1, plan_actions, 3, padding = 1, bias = False)

        self.action_selector = ActionSelector(
            num_plans = num_plans,
            soft_maxpool = soft_maxpool,
            soft_maxpool_temperature = soft_maxpool_temperature
        )

        # value iteration network

        value_iteration_module = ValueIteration(
            num_actions,
            num_plans = num_plans,
            soft_maxpool = soft_maxpool,
            soft_maxpool_temperature = soft_maxpool_temperature,
            **vi_module_kwargs
        )

        self.planner = ValueIterationNetwork(
            value_iteration_module,
            depth = depth,
            **vin_kwargs
        )

        self.final_cropout_kernel_size = final_cropout_kernel_size

        final_value_dim = final_cropout_kernel_size ** 2
        final_state_dim = state_dim * final_cropout_kernel_size ** 2

        # final attention across all values

        self.attn_pool = Attention(
            dim = final_value_dim + final_state_dim,
            dim_out = final_value_dim,
            causal = True,
            dim_head = attn_dim_head,
            **attn_kwargs
        )

        self.rotary_pos_emb = RotaryEmbedding(attn_dim_head)

        # losses

        self.to_action_logits = nn.Linear(num_plans * final_value_dim, num_actions, bias = False)

        self.loss_every_num_layers = loss_every_num_layers

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(
        self,
        state,
        reward,
        agent_positions,
        target_actions = None,
        depth = None
    ):
        batch, num_plans = state.shape[0], self.num_plans

        depth = default(depth, self.depth)

        state_reward, _ = pack([state, reward], 'b * h w')

        reward = self.reward_mapper(state_reward)

        q_values = self.to_init_value(reward)

        init_value = self.action_selector(q_values)

        layer_values = self.planner(init_value, reward, depth = depth)

        # values across all layers

        layer_values = torch.stack(layer_values)
        layer_values = torch.flip(layer_values, dims = (0,)) # so depth goes from last layer to first

        layer_values, inverse_pack = pack_one(layer_values, '* c h w')  # pack the depth with the batch and calculate loss across all layers in one go

        # unfold the values across all layers, and select out the coordinate for the agent position

        unfolded_layer_values = F.unfold(layer_values, self.final_cropout_kernel_size, padding = self.final_cropout_kernel_size // 2)

        unfolded_layer_values = inverse_pack(unfolded_layer_values, '* a hw')

        # get only the values at the agent coordinates

        height_width_strides = tensor(state.stride()[-2:])
        agent_position_hw_index = (agent_positions * height_width_strides).sum(dim = -1)

        unfolded_layer_values = einx.get_at('d b v [hw], b -> b d v', unfolded_layer_values, agent_position_hw_index)

        unfolded_layer_values = rearrange(unfolded_layer_values, 'b d (p v) -> (b p) d v', p = num_plans)

        # concat states onto each value and do an attention across all values across all layers

        unfolded_state_values = F.unfold(state, self.final_cropout_kernel_size, padding = self.final_cropout_kernel_size // 2)
        unfolded_state_values = einx.get_at('b s [hw], b -> b s', unfolded_state_values, agent_position_hw_index)

        unfolded_state_values = repeat(unfolded_state_values, 'b s -> (b p) d s', d = depth, p = num_plans)

        state_and_all_values, _ = pack([unfolded_layer_values, unfolded_state_values], 'b d *')

        # allow each layer values to attend to all layers of the past
        # if there are multiple plans, layer values of each plan will have its own attention map

        rotary_pos_emb = self.rotary_pos_emb(torch.arange(depth, device = self.device))

        attended = self.attn_pool(state_and_all_values, rotary_pos_emb = rotary_pos_emb)

        # add the output of the 'causal' attention (across depth) to the values

        unfolded_layer_values = unfolded_layer_values + attended

        # fold num plans dimension back into action dimension

        unfolded_layer_values = rearrange(unfolded_layer_values, '(b p) d v -> b d (p v)', p = num_plans)

        # gather all layers for calculating losses

        if self.loss_every_num_layers < 1:
            # anything less than 1, just treat as only loss on last layer
            unfolded_layer_values = unfolded_layer_values[:, :1]
        else:
            unfolded_layer_values = unfolded_layer_values[:, ::self.loss_every_num_layers]

        num_layers_calc_loss = unfolded_layer_values.shape[1]

        # calculating action logits

        action_logits = self.to_action_logits(unfolded_layer_values)

        # return logits if no labels

        if not exists(target_actions):
            return action_logits

        # else calculate the loss

        losses = F.cross_entropy(
            rearrange(action_logits, 'b d a -> b a d'),
            repeat(target_actions, 'b -> b d', d = num_layers_calc_loss),
            reduction = 'none'
        )

        losses = reduce(losses, 'b d -> b', 'sum') # sum losses across depth, could do some sort of weighting too

        return losses.mean()

==> scaling_vin_pytorch-0.0.6-py3-none-any.whl: scaling_vin_pytorch/vin.py <==
from functools import partial, wraps

import torch
import torch.nn.functional as F
from torch.nn import Module, ModuleList
from torch import nn, tensor, Tensor, is_tensor
from torch.utils.checkpoint import checkpoint_sequential

from x_transformers import Attention
from x_transformers.x_transformers import RotaryEmbedding

import einx
from einops import rearrange, repeat, reduce, einsum, pack, unpack

# ein notation
# b - batch
# c - channels
# s - state dimension
# v - values
# a - actions (channels)
# o - output (channels)
# i - input (channels)
# h - height
# w - width
# d - depth of value iteration network
# p - number of plans (different value maps transitioned on)

# helpers

def exists(v):
    return v is not None

def default(v, d):
    return v if exists(v) else d

def divisible_by(num, den):
    return (num % den) == 0

def is_odd(n):
    return not divisible_by(n, 2)

def pack_one(t, pattern):
    output, packed_shape = pack([t], pattern)

    def inverse_fn(packed, override_pattern = None):
        override_pattern = default(override_pattern, pattern)

        return unpack(packed, packed_shape, override_pattern)[0]

    return output, inverse_fn

def should_checkpoint(
    self,
    inputs: Tensor | tuple[Tensor, ...],
    check_instance_variable = 'checkpoint'
) -> bool:
    if is_tensor(inputs):
        inputs = (inputs,)

    return (
        self.training and
        any([i.requires_grad for i in inputs]) and
        (not exists(check_instance_variable) or getattr(self, check_instance_variable, False))
    )

# tensor helpers

def soft_maxpool(
    t,
    temperature = 1.,
    dim = 1,
):
    t = t / temperature
    out = t.softmax(dim = dim) * t
    out = out.sum(dim = dim)
    return out * temperature

# modules and classes

class ActionSelector(Module):
    def __init__(
        self,
        num_plans = 1,
        soft_maxpool = False,
        soft_maxpool_temperature = 1,
    ):
        super().__init__()
        self.num_plans = num_plans

        # allow for softmax(x) * x pooling
        # https://mpflueger.github.io/assets/pdf/svin_iclr2018_v2.pdf
        self.soft_maxpool = soft_maxpool
        self.soft_maxpool_temperature = soft_maxpool_temperature

    def forward(self, q_values):

        q_values = rearrange(q_values, 'b (p a) ... -> b p a ...', p = self.num_plans)

        if not self.soft_maxpool:
            next_values = q_values.amax(dim = 2)
        else:
            next_values = soft_maxpool(q_values, temperature = self.soft_maxpool_temperature, dim = 2)

        return next_values

class ValueIteration(Module):
    def __init__(
        self,
        action_channels,
        *,
        num_plans = 1,
        receptive_field = 3,
        pad_value = 0.,
        soft_maxpool = False,
        soft_maxpool_temperature = 1.,
        dynamic_transition_kernel = True,
    ):
        super().__init__()
        assert is_odd(receptive_field)
        padding = receptive_field // 2

        self.action_channels = action_channels
        plan_actions = num_plans * action_channels
        self.num_plans = num_plans

        conv_out = action_channels * (num_plans ** 2) * (1 if not dynamic_transition_kernel else receptive_field ** 2)

        self.transition = nn.Conv2d(num_plans, conv_out, receptive_field, padding = padding, bias = False)

        self.kernel_size = receptive_field

        self.pad_value = pad_value
        self.pad = partial(F.pad, pad = (padding,) * 4, value = pad_value)

        self.padding = padding

        self.action_selector = ActionSelector(
            num_plans = num_plans,
            soft_maxpool = soft_maxpool,
            soft_maxpool_temperature = soft_maxpool_temperature
        )

        self.dynamic_transition_kernel = dynamic_transition_kernel

    def forward(
        self,
        values,
        rewards
    ):
        pad = self.pad

        rewards_and_values = values + rewards

        # prepare for transition

        transition_weight = self.transition.weight

        # dynamic transition kernel - in other words, first convolution outputs the transition kernel

        if self.dynamic_transition_kernel:

            dynamic_transitions = F.conv2d(pad(rewards_and_values), transition_weight)

            # reshape the output into the next transition weight kernel

            dynamic_transitions = rearrange(dynamic_transitions, 'b (o i k1 k2) h w -> b o h w (i k1 k2)', k1 = self.kernel_size, k2 = self.kernel_size, i = self.num_plans)

            # the author then uses a softmax on the dynamic transition kernel
            # this actually becomes form of attention pooling seen in other papers

            dynamic_transitions = F.softmax(dynamic_transitions, dim = -1)

            # unfold the reward and values to manually do "conv" with data dependent kernel

            width = rewards_and_values.shape[-1] # for rearranging back after unfold

            unfolded_values = F.unfold(pad(rewards_and_values), self.kernel_size)
            unfolded_values = rearrange(unfolded_values, 'b i (h w) -> b i h w', w = width)

            # dynamic kernel

            q_values = einsum(unfolded_values, dynamic_transitions, 'b i h w, b o h w i -> b o h w')

        else:
            # transition

            q_values = F.conv2d(pad(rewards_and_values), transition_weight)

        # selecting the next action

        next_values = self.action_selector(q_values)

        return next_values

class ValueIterationNetwork(Module):
    def __init__(
        self,
        vi_module: ValueIteration,
        depth,
        checkpoint = True,
        checkpoint_segments = None
    ):
        super().__init__()
        self.depth = depth
        self.vi_module = vi_module

        self.checkpoint = checkpoint
        self.checkpoint_segments = checkpoint_segments

    def forward(
        self,
        values,
        rewards,
        depth = None
    ):
        depth = default(depth, self.depth)

        values, _ = pack_one(values, 'b * h w')
        rewards, _ = pack_one(rewards, 'b * h w')

        # checkpointable or not

        if depth > 1 and should_checkpoint(self, (values, rewards)):

            layer_values = [None] * depth

            segments = default(self.checkpoint_segments, depth)
            checkpoint_fn = partial(checkpoint_sequential, segments = segments, use_reentrant = False)

            def recurrent_layer(inputs):
                layer_ind, values, rewards, *layer_values = inputs

                next_values = self.vi_module(values, rewards)

                layer_values[layer_ind] = next_values

                return layer_ind + 1, next_values, rewards, *layer_values

            all_recurrent_layers = (recurrent_layer,) * depth

            _, _, _, *layer_values = checkpoint_fn(all_recurrent_layers, input = (0, values, rewards, *layer_values))

        else:

            layer_values = []

            for _ in range(depth):
                values = self.vi_module(values, rewards)
                layer_values.append(values)

        # return all values

        assert len(layer_values) == depth

        return layer_values

# main class

class ScalableVIN(Module):
    def __init__(
        self,
        state_dim,
        reward_dim,
        num_actions,
        num_plans = 1,
        dim_hidden = 150,
        init_kernel_size = 7,
        depth = 100,                    # they scaled this to 5000 in the paper to solve 100x100 maze
        loss_every_num_layers = 4,      # calculating loss every so many layers in the value iteration network
        final_cropout_kernel_size = 3,
        soft_maxpool = False,
        soft_maxpool_temperature = 1.,
        vi_module_kwargs: dict = dict(),
        vin_kwargs: dict = dict(),
        attn_dim_head = 64,
        attn_kwargs: dict = dict()
    ):
        super().__init__()
        self.depth = depth
        self.num_actions = num_actions
        self.num_plans = num_plans

        plan_actions = num_actions * num_plans

        self.reward_mapper = nn.Sequential(
            nn.Conv2d(state_dim + reward_dim, dim_hidden, 3, padding = 1),
            nn.Conv2d(dim_hidden, 1, 1, bias = False)
        )

        self.to_init_value = nn.Conv2d(1, plan_actions, 3, padding = 1, bias = False)

        self.action_selector = ActionSelector(
            num_plans = num_plans,
            soft_maxpool = soft_maxpool,
            soft_maxpool_temperature = soft_maxpool_temperature
        )

        # value iteration network

        value_iteration_module = ValueIteration(
            num_actions,
            num_plans = num_plans,
            soft_maxpool = soft_maxpool,
            soft_maxpool_temperature = soft_maxpool_temperature,
            **vi_module_kwargs
        )

        self.planner = ValueIterationNetwork(
            value_iteration_module,
            depth = depth,
            **vin_kwargs
        )

        self.final_cropout_kernel_size = final_cropout_kernel_size

        final_value_dim = final_cropout_kernel_size ** 2
        final_state_dim = state_dim * final_cropout_kernel_size ** 2

        # final attention across all values

        self.attn_pool = Attention(
            dim = final_value_dim + final_state_dim,
            dim_out = final_value_dim,
            causal = True,
            dim_head = attn_dim_head,
            **attn_kwargs
        )

        self.rotary_pos_emb = RotaryEmbedding(attn_dim_head)

        # losses

        self.to_action_logits = nn.Linear(num_plans * final_value_dim, num_actions, bias = False)

        self.loss_every_num_layers = loss_every_num_layers

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(
        self,
        state,
        reward,
        agent_positions,
        target_actions = None,
        depth = None
    ):
        batch, num_plans = state.shape[0], self.num_plans

        depth = default(depth, self.depth)

        state_reward, _ = pack([state, reward], 'b * h w')

        reward = self.reward_mapper(state_reward)

        q_values = self.to_init_value(reward)

        init_value = self.action_selector(q_values)

        layer_values = self.planner(init_value, reward, depth = depth)

        # values across all layers

        layer_values = torch.stack(layer_values)
        layer_values = torch.flip(layer_values, dims = (0,)) # so depth goes from last layer to first

        layer_values, inverse_pack = pack_one(layer_values, '* c h w')  # pack the depth with the batch and calculate loss across all layers in one go

        # unfold the values across all layers, and select out the coordinate for the agent position

        unfolded_layer_values = F.unfold(layer_values, self.final_cropout_kernel_size, padding = self.final_cropout_kernel_size // 2)

        unfolded_layer_values = inverse_pack(unfolded_layer_values, '* a hw')

        # get only the values at the agent coordinates

        height_width_strides = tensor(state.stride()[-2:])
        agent_position_hw_index = (agent_positions * height_width_strides).sum(dim = -1)

        unfolded_layer_values = einx.get_at('d b v [hw], b -> b d v', unfolded_layer_values, agent_position_hw_index)

        unfolded_layer_values = rearrange(unfolded_layer_values, 'b d (p v) -> (b p) d v', p = num_plans)

        # concat states onto each value and do an attention across all values across all layers

        unfolded_state_values = F.unfold(state, self.final_cropout_kernel_size, padding = self.final_cropout_kernel_size // 2)
        unfolded_state_values = einx.get_at('b s [hw], b -> b s', unfolded_state_values, agent_position_hw_index)

        unfolded_state_values = repeat(unfolded_state_values, 'b s -> (b p) d s', d = depth, p = num_plans)

        state_and_all_values, _ = pack([unfolded_layer_values, unfolded_state_values], 'b d *')

        # allow each layer values to attend to all layers of the past
        # if there are multiple plans, layer values of each plan will have its own attention map

        rotary_pos_emb = self.rotary_pos_emb(torch.arange(depth, device = self.device))

        attended = self.attn_pool(state_and_all_values, rotary_pos_emb = rotary_pos_emb)

        # add the output of the 'causal' attention (across depth) to the values

        unfolded_layer_values = unfolded_layer_values + attended

        # fold num plans dimension back into action dimension

        unfolded_layer_values = rearrange(unfolded_layer_values, '(b p) d v -> b d (p v)', p = num_plans)

        # gather all layers for calculating losses

        if self.loss_every_num_layers < 1:
            # anything less than 1, just treat as only loss on last layer
            unfolded_layer_values = unfolded_layer_values[:, :1]
        else:
            unfolded_layer_values = unfolded_layer_values[:, ::self.loss_every_num_layers]

        num_layers_calc_loss = unfolded_layer_values.shape[1]

        # calculating action logits

        action_logits = self.to_action_logits(unfolded_layer_values)

        # return logits if no labels

        if not exists(target_actions):
            return action_logits

        # else calculate the loss

        losses = F.cross_entropy(
            rearrange(action_logits, 'b d a -> b a d'),
            repeat(target_actions, 'b -> b d', d = num_layers_calc_loss),
            reduction = 'none'
        )

        losses = reduce(losses, 'b d -> b', 'sum') # sum losses across depth, could do some sort of weighting too

        return losses.mean()
